# parse_options: {"target-version": "3.14"}
def f() -> (y := 3): ...
def g(arg: (x := 1)): ...
def outer():
    def i(x: (yield 1)): ...
    def k() -> (yield 1): ...
    def m(x: (yield from 1)): ...
    def o() -> (yield from 1): ...
async def outer():
    def f() -> (await 1): ...
    def g(arg: (await 1)): ...
